SystemOrganization addCategory: #MetaInterpreter!

TestCase subclass: #TestInterpreter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/8/2013 16:13'!
passAsArg: t

! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/2/2013 16:40'!
testDerivativeHooksWithHalt

	"press proceed for the test to finish"

	|t1 t2|
	
	"register halt on creation"
	Interpreter onMessageReceivedForObject: TestObject do: [:reifications :reflectogram | 
		reifications message selector = #new ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: TestObject.
	
	t1 := TestObject new. "should halt"
	t2 := TestObject new. "should halt."
	
	Interpreter removeMetaBehaviorOf: TestObject.
	
	"register halt on object in invoke -- this can be done either by registering the argument or the receiver"
	Interpreter onArgumentForObject: t2 do: [:reifications :reflectogram |
			reifications receiver = t1 ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: t2.
	
	t1 testPassingArgument: t2. "should halt"

	Interpreter removeMetaBehaviorOf: t2.
	
	"register halt on interaction"	
	Interpreter onMessageSendFromObject: t1 do: [:reifications :override | 
		reifications receiver = t2 ifTrue: [self halt].
	].

	Interpreter onMessageSendFromObject: t2 do: [:reifications :override | 
		reifications receiver = t1 ifTrue: [self halt].
	].

	Interpreter enableMetaBehaviorFor: t1.
	Interpreter enableMetaBehaviorFor: t2.

	t1 testInteraction: t2. "should halt because t1 invokes a method of t2"
	t2 testInteraction: t1. "should halt because t2 invokes a method of t1"
	
	Interpreter disableMetaBehaviorFor: t1.
	Interpreter disableMetaBehaviorFor: t2.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/2/2013 16:50'!
testHookArg

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter onArgumentForObject: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/2/2013 16:50'!
testHookRead

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter onIvReadForObject: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.			
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: ivReadCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/2/2013 16:51'!
testHookReceive

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter onMessageReceivedForObject: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.		
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/2/2013 16:51'!
testHookSend

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
	
		Interpreter onMessageSendFromObject: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.	
		].
	

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: sendCounter = (2*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/2/2013 16:51'!
testHookWrite

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	

		Interpreter onIvWriteForObject: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.		
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: ivWriteCounter = (1*100).
! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/2/2013 16:52'!
testHooks

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter onMessageReceivedForObject: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.		
		].
	
		Interpreter onMessageSendFromObject: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.	
		].
	
		Interpreter onIvReadForObject: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.			
		].

		Interpreter onIvWriteForObject: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.		
		].

		Interpreter onArgumentForObject: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
	self assert: sendCounter = (2*100).
	self assert: ivReadCounter = (1*100).
	self assert: ivWriteCounter = (1*100).
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/2/2013 16:53'!
testHooksWithHalt

	"press proceed for the test to finish"

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

		
	t := TestObject new.
	
	Interpreter onMessageReceivedForObject: t do: [:reifications :reflectogram | 
		self halt.
		receiveCounter := receiveCounter + 1.		
	].
	
	Interpreter onMessageSendFromObject: t do: [:reifications :reflectogram | 
		self halt.
		sendCounter := sendCounter + 1.	
	].
	
	Interpreter onIvReadForObject: t do: [:reifications :reflectogram | 
		self halt.
		ivReadCounter := ivReadCounter + 1.			
	].

	Interpreter onIvWriteForObject: t do: [:reifications :reflectogram | 
		self halt.
		ivWriteCounter := ivWriteCounter + 1.		
	].

	Interpreter onArgumentForObject: t do: [:reifications :reflectogram |
		self halt.
		argCounter := argCounter + 1
	].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 5). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	
	self assert: receiveCounter = 3.
	self assert: sendCounter = 2.
	self assert: ivReadCounter = 1.
	self assert: ivWriteCounter = 1.
	self assert: argCounter = 1.! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/2/2013 17:19'!
testHooksWithOverride

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.
	sendCounter := 0.
	ivReadCounter := 0.
	ivWriteCounter := 0.
	argCounter := 0.

	1 to: 100 do: [:i| 	
		
		t := TestObject new.
	
		Interpreter onMessageReceivedForObject: t do: [:reifications :reflectogram | 
			receiveCounter := receiveCounter + 1.		
		].
	
		Interpreter onMessageSendFromObject: t do: [:reifications :reflectogram | 
			sendCounter := sendCounter + 1.	
		].
	
		Interpreter onIvReadForObject: t do: [:reifications :reflectogram | 
			ivReadCounter := ivReadCounter + 1.
			reflectogram replace. "this callback will override default behavior"
			666. "return value of iv read"	
		].

		Interpreter onIvWriteForObject: t do: [:reifications :reflectogram | 
			ivWriteCounter := ivWriteCounter + 1.		
		].

		Interpreter onArgumentForObject: t do: [:reifications :reflectogram |
			argCounter := argCounter + 1
		].

		Interpreter enableMetaBehaviorFor: t.
	
		self passAsArg: t. "should trigger meta arg"
		t x: 5. "should trigger meta receive and meta-write".
		self assert: (t x = 666). "should trigger meta-receive and meta-read".
		t testSenderHook. "should trigger 2 meta-sends"
	
		"Interpreter disableMetaBehaviorFor: t."
		Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	].
	
	self assert: receiveCounter = (3*100).
	self assert: sendCounter = (2*100).
	self assert: ivReadCounter = (1*100).
	self assert: ivWriteCounter = (1*100).
	self assert: argCounter = (1*100).! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/2/2013 17:48'!
testOneShot

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.

	t := TestObject new.

	Interpreter onMessageReceivedForObject: t do: [:reifications :reflectogram | 
		receiveCounter := receiveCounter + 1.
		reflectogram beOneShot.
	].


	Interpreter enableMetaBehaviorFor: t.

	1 to: 100 do: [:i| 	
			
		t class.	
	
	].
	
	"Interpreter disableMetaBehaviorFor: t."
	Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	self assert: receiveCounter = 1. "it was one shot"! !

!TestInterpreter methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/2/2013 17:49'!
testOneShotWithPost

	|t receiveCounter sendCounter ivReadCounter ivWriteCounter argCounter|
	
	receiveCounter := 0.

	t := TestObject new.

	Interpreter onMessageReceivedForObject: t do: [:reifications :reflectogram | 
		receiveCounter := receiveCounter + 1.
		reflectogram beOneShot.
		reflectogram postAction: [receiveCounter := receiveCounter + 1].
	].


	Interpreter enableMetaBehaviorFor: t.

	1 to: 100 do: [:i| 	
			
		t class.	
	
	].
	
	"Interpreter disableMetaBehaviorFor: t."
	Interpreter removeMetaBehaviorOf: t. "also removes low-level meta-objects"
	
	self assert: receiveCounter = 2. "it was one shot"! !

Object subclass: #BenchmarkObject
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!BenchmarkObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 14:41'!
test
	| a b c |
	a := 1.
	b := 2.
	c := a + b.
	^ c
! !

Object subclass: #Interpreter
	instanceVariableNames: ''
	classVariableNames: 'BaseMetaDict'
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Interpreter class methodsFor: 'initialiazation' stamp: 'NickPapoulias 5/8/2013 15:04'!
baseMetaDict

	^ BaseMetaDict! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 16:12'!
disableMetaBehaviorFor: anObject

	"unbind"
	
	self primitiveSetExperimentalBitOf: anObject to: false. ! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 16:11'!
enableMetaBehaviorFor: anObject

	"bind"
	
	self primitiveSetExperimentalBitOf: anObject to: true. ! !

!Interpreter class methodsFor: 'initialiazation' stamp: 'NickPapoulias 6/2/2013 15:59'!
initialize

	super initialize.
	BaseMetaDict := FLLargeIdentityDictionary new.
	Smalltalk specialObjectsArray becomeForward: self newSpecialObjectsArray.! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 16:13'!
isMetaBehaviorEnabledFor: anObject


	^ self primitiveGetExperimentalBitOf: anObject! !

!Interpreter class methodsFor: 'initialiazation' stamp: 'NickPapoulias 5/29/2013 14:17'!
newSpecialObjectsArray
	"Smalltalk recreateSpecialObjectsArray"
	
	"To external package developers:
	**** DO NOT OVERRIDE THIS METHOD.  *****
	If you are writing a plugin and need additional special object(s) for your own use, 
	use addGCRoot() function and use own, separate special objects registry "
	
	"The Special Objects Array is an array of objects used by the Squeak virtual machine.
	 Its contents are critical and accesses to it by the VM are unchecked, so don't even
	 think of playing here unless you know what you are doing."
	
	| newArray |
	newArray := Array new: 76.
	"Nil false and true get used throughout the interpreter"
	newArray at: 1 put: nil.
	newArray at: 2 put: false.
	newArray at: 3 put: true.
	"This association holds the active process (a ProcessScheduler)"
	newArray at: 4 put: (Smalltalk globals associationAt: #Processor).
	"Numerous classes below used for type checking and instantiation"
	newArray at: 5 put: Bitmap.
	newArray at: 6 put: SmallInteger.
	newArray at: 7 put: ByteString.
	newArray at: 8 put: Array.
	newArray at: 9 put: Smalltalk.
	newArray at: 10 put: Float.
	newArray at: 11 put: MethodContext.
	newArray at: 12 put: nil. "Was BlockContext"
	newArray at: 13 put: Point.
	newArray at: 14 put: LargePositiveInteger.
	newArray at: 15 put: Display.
	newArray at: 16 put: Message.
	newArray at: 17 put: CompiledMethod.
	newArray at: 18 put: (Smalltalk specialObjectsArray at: 18).
	"(low space Semaphore)"
	newArray at: 19 put: Semaphore.
	newArray at: 20 put: Character.
	newArray at: 21 put: #doesNotUnderstand:.
	newArray at: 22 put: #cannotReturn:.
	newArray at: 23 put: nil. "This is the process signalling low space."
	"An array of the 32 selectors that are compiled as special bytecodes,
	 paired alternately with the number of arguments each takes."
	newArray at: 24 put: #(	#+ 1 #- 1 #< 1 #> 1 #<= 1 #>= 1 #= 1 #~= 1
							#* 1 #/ 1 #\\ 1 #@ 1 #bitShift: 1 #// 1 #bitAnd: 1 #bitOr: 1
							#at: 1 #at:put: 2 #size 0 #next 0 #nextPut: 1 #atEnd 0 #== 1 nil 0
							nil 1 #value 0 #value: 1 #do: 1 #new 0 #new: 1 #x 0 #y 0 ).
	"An array of the 255 Characters in ascii order.
	 Cog inlines table into machine code at: prim so do not regenerate it."
	newArray at: 25 put: (Smalltalk specialObjectsArray at: 25).
	newArray at: 26 put: #mustBeBoolean.
	newArray at: 27 put: ByteArray.
	newArray at: 28 put: Process.
	"An array of up to 31 classes whose instances will have compact headers"
	newArray at: 29 put: Smalltalk compactClassesArray.
	newArray at: 30 put: (Smalltalk specialObjectsArray at: 30).
	"(delay Semaphore)"
	newArray at: 31 put: (Smalltalk specialObjectsArray at: 31).
	"(user interrupt Semaphore)"
	"Prototype instances that can be copied for fast initialization"
	newArray at: 32 put: (Float new: 2).
	newArray at: 33 put: (LargePositiveInteger new: 4).
	newArray at: 34 put: 0@0.
	newArray at: 35 put: #cannotInterpret:.
	"Note: This must be fixed once we start using context prototypes (yeah, right)"
	"(MethodContext new: CompiledMethod fullFrameSize)."
	newArray at: 36 put: (Smalltalk specialObjectsArray at: 36). "Is the prototype MethodContext (unused by the VM)"
	newArray at: 37 put: BlockClosure.
	"(BlockContext new: CompiledMethod fullFrameSize)."
	newArray at: 38 put: (Smalltalk specialObjectsArray at: 38). "Is the prototype BlockContext (unused by the VM)"
	"array of objects referred to by external code"
	newArray at: 39 put: (Smalltalk specialObjectsArray at: 39).	"preserve external semaphores"
	newArray at: 40 put: nil. "Reserved for Mutex in Cog VMs"
	newArray at: 41 put: nil. "Reserved for a LinkedList instance for overlapped calls in CogMT"
	"finalization Semaphore"
	newArray at: 42 put: ((Smalltalk specialObjectsArray at: 42) ifNil: [Semaphore new]).
	newArray at: 43 put: LargeNegativeInteger.
	"External objects for callout.
	 Note: Written so that one can actually completely remove the FFI."
	newArray at: 44 put: (Smalltalk at: #ExternalAddress ifAbsent: []).
	newArray at: 45 put: (Smalltalk at: #ExternalStructure ifAbsent: []).
	newArray at: 46 put: (Smalltalk at: #ExternalData ifAbsent: []).
	newArray at: 47 put: (Smalltalk at: #ExternalFunction ifAbsent: []).
	newArray at: 48 put: (Smalltalk at: #ExternalLibrary ifAbsent: []).
	newArray at: 49 put: #aboutToReturn:through:.
	newArray at: 50 put: #run:with:in:.
	"51 reserved for immutability message"
	"newArray at: 51 put: #attemptToAssign:withIndex:."
	newArray at: 52 put: #(nil "nil => generic error" #'bad receiver'
							#'bad argument' #'bad index'
							#'bad number of arguments'
							#'inappropriate operation'  #'unsupported operation'
							#'no modification' #'insufficient object memory'
							#'insufficient C memory' #'not found' #'bad method'
							#'internal error in named primitive machinery'
							#'object may move').
	"53 to 55 are for Alien"
	newArray at: 53 put: (Smalltalk at: #Alien ifAbsent: []).
	newArray at: 54 put: #invokeCallback:stack:registers:jmpbuf:.
	newArray at: 55 put: (Smalltalk at: #UnsafeAlien ifAbsent: []).

	"Weak reference finalization"
	newArray at: 56 put: (Smalltalk at: #WeakFinalizationList ifAbsent: []).
	newArray at: 59 put: #metaReceived:.
	newArray at: 60 put: #metaRef.
	newArray at: 61 put: #metaReadAt:.
	newArray at: 62 put: #metaWriteAt:Put:.
	newArray at: 63 put: #metaArgIn:for:.
	newArray at: 64 put: Interpreter.
	newArray at: 65 put: #metaSend:from:.
	newArray at: 66 put: Interpreter baseMetaDict.
	"optimization"
	newArray at: 67 put: Interpreter class>>#object:metaReceived:.
	newArray at: 68 put: Interpreter class>>#object:metaSend:withReceiver:.
	newArray at: 69 put: Interpreter class>>#object:metaReadAt:.
	newArray at: 70 put: Interpreter class>>#object:metaWriteAt:Put:.
	newArray at: 71 put: Interpreter class>>#object:metaArgIn:withReceiver:.
	
	newArray at: 72 put: #object:metaReceived:.
	newArray at: 73 put: #object:metaSend:withReceiver:.
	newArray at: 74 put: #object:metaReadAt:.
	newArray at: 75 put: #object:metaWriteAt:Put:.
	newArray at: 76 put: #object:metaArgIn:withReceiver:.
	^newArray! !

!Interpreter class methodsFor: 'introspection' stamp: 'NickPapoulias 5/5/2013 18:43'!
object: anObject at: anIndex

	<primitive: 73>
	! !

!Interpreter class methodsFor: 'modification' stamp: 'NickPapoulias 5/5/2013 18:43'!
object: anObject at: anIndex put: anotherObject

	<primitive: 74>
	! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/3/2013 12:21'!
object: anObject metaArgIn: aMessage withReceiver: rcvr

	| implicitMetaObject defaultAction reflectogram |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	defaultAction := [Interpreter object: rcvr perform: aMessage selector withArguments: aMessage arguments].

	implicitMetaObject argCallback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ defaultAction value
	]  ifNotNil:  [
		implicitMetaObject argCallback valueWithArguments: { 
			Reifications with: thisContext . 
			reflectogram := Reflectogram for: anObject withDefaultAction: defaultAction}. 
	].
	
	^ reflectogram returnValue ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/3/2013 12:32'!
object: anObject metaReadAt: anIndex

	| implicitMetaObject returnValue defaultAction reflectogram |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	defaultAction := [Interpreter object: anObject at: anIndex].

	implicitMetaObject readCallback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ defaultAction value
	]  ifNotNil:  [
		implicitMetaObject readCallback valueWithArguments: { 
			Reifications with: thisContext . 
			reflectogram := Reflectogram for: anObject withDefaultAction: defaultAction}. 
	].
	
	^ reflectogram returnValue 
! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/3/2013 12:31'!
object: anObject metaReceived: aMessage

	| implicitMetaObject returnValue defaultAction reflectogram |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	defaultAction := [ Interpreter object: anObject perform: aMessage selector withArguments: aMessage arguments].

	implicitMetaObject receiveCallback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ defaultAction value
	]  ifNotNil:  [
		implicitMetaObject receiveCallback valueWithArguments: { 
			Reifications with: thisContext . 
			reflectogram := Reflectogram for: anObject withDefaultAction: defaultAction}. 
	].
	
	^ reflectogram returnValue ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/3/2013 12:32'!
object: anObject metaSend: aMessage withReceiver: rcvr

	| implicitMetaObject returnValue defaultAction reflectogram |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	defaultAction := [Interpreter object: rcvr  perform: aMessage selector withArguments: aMessage arguments].

	implicitMetaObject sendCallback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ defaultAction value
	]  ifNotNil:  [
		implicitMetaObject sendCallback valueWithArguments: { 
			Reifications with: thisContext . 
			reflectogram := Reflectogram for: anObject withDefaultAction: defaultAction}. 
	].
	
	^ reflectogram returnValue ! !

!Interpreter class methodsFor: 'intercession' stamp: 'NickPapoulias 6/3/2013 12:33'!
object: anObject metaWriteAt: anIndex Put: aValue

	| implicitMetaObject returnValue defaultAction reflectogram |
	"Transcript show: thisContext; show: aMessage ; cr." 	

	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	defaultAction := [ Interpreter object: anObject at: anIndex put: aValue].

	implicitMetaObject writeCallback ifNil: [
		Interpreter enableMetaBehaviorFor: anObject.
		^ defaultAction value
	]  ifNotNil:  [
		implicitMetaObject writeCallback valueWithArguments: { 
			Reifications with: thisContext . 
			reflectogram := Reflectogram for: anObject withDefaultAction: defaultAction}. 
	].
	
	^ reflectogram returnValue ! !

!Interpreter class methodsFor: 'execution' stamp: 'NickPapoulias 5/5/2013 18:44'!
object: anObject perform: aSelector withArguments: args 

	<primitive: 84>! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onArgumentForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject argCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onIvReadForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject readCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onIvWriteForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject writeCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onMessageReceivedForObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject receiveCallback: aCallback.
! !

!Interpreter class methodsFor: 'meta-callback-registration' stamp: 'NickPapoulias 5/8/2013 15:39'!
onMessageSendFromObject: anObject do: aCallback

	| implicitMetaObject |
	
	implicitMetaObject := BaseMetaDict at: anObject ifAbsent: [
		BaseMetaDict at: anObject put: LowLevelImplicitMetaObject new.	
	].

	implicitMetaObject sendCallback: aCallback.
! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 15:47'!
primitiveGetExperimentalBitOf: anObject

	<primitive: 150>
	self primitiveFailed! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 5/7/2013 15:22'!
primitiveSetExperimentalBitOf: anObject to: aBoolean

	<primitive: 151>
	self primitiveFailed! !

!Interpreter class methodsFor: 'binding' stamp: 'NickPapoulias 6/2/2013 16:02'!
removeMetaBehaviorOf: anObject

	"unbind and remove implicit meta-object"
	self disableMetaBehaviorFor: anObject.
	BaseMetaDict removeKey: anObject ifAbsent: [].! !

Object subclass: #LazyReifications
	instanceVariableNames: 'interpreterContext'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!LazyReifications class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 10:54'!
with: anInterpreterContext

	^ self new interpreterContext: anInterpreterContext! !

!LazyReifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:10'!
interpreterContext

	^ interpreterContext! !

!LazyReifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:10'!
interpreterContext: aContext

	interpreterContext := aContext! !

LazyReifications subclass: #Reifications
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:11'!
context

	^ interpreterContext sender ! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:13'!
index

	^ interpreterContext tempNamed: #anIndex! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
message

	^ self interpreterContext tempNamed: #aMessage
! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
method

	^ self context method! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:13'!
receiver

	^ interpreterContext tempNamed: #rcvr
! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
sender

	^ self context receiver! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 11:08'!
sourceRange

	^ self context debuggerMap 
						rangeForPC: self context pc
						contextIsActiveContext: false 
! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:13'!
trigger

	^ interpreterContext tempNamed: #anObject! !

!Reifications methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:14'!
value

	^ interpreterContext tempNamed: #aValue! !

Object subclass: #LowLevelImplicitMetaObject
	instanceVariableNames: 'receiveCallback sendCallback readCallback writeCallback argCallback'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!LowLevelImplicitMetaObject class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/7/2013 17:57'!
new

	^ self basicNew initialize! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
argCallback
	^ argCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
argCallback: anObject
	argCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 6/2/2013 15:16'!
initialize 

	super initialize.

	"self receiveCallback: [:reifications :override | ].

	self sendCallback: [:reifications :override | ].

	self readCallback: [:reifications :override | ].

	self writeCallback: [:reifications :override | ].
	
	self	 argCallback: [:reifications :override | ].
		"! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
readCallback
	^ readCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
readCallback: anObject
	readCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 14:58'!
receiveCallback
	^ receiveCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 14:58'!
receiveCallback: anObject
	receiveCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
sendCallback
	^ sendCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
sendCallback: anObject
	sendCallback := anObject! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
writeCallback
	^ writeCallback! !

!LowLevelImplicitMetaObject methodsFor: 'accessing' stamp: 'NickPapoulias 5/7/2013 11:48'!
writeCallback: anObject
	writeCallback := anObject! !

Object subclass: #MetaBenchmark
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 14:36'!
benchmark: aBlock times: anInteger
	| runs average deviation |

	runs := [ 
		(1 to: anInteger) 
			collect: [ :each | aBlock timeToRun ] ]
		valueUnpreemptively.
	average := runs average asFloat.
	deviation := ((runs 
		detectSum: [ :each | (each - average) * (each - average) ]) 
			/ runs size) sqrt.
	^ Array with: average with: deviation! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/10/2013 14:41'!
message
	"self new message" 
	
	|  object  |
	object := BenchmarkObject new.
	^ self benchmark: [ 1 to: 1000000 do: [ :each | object test ] ] times: 100.! !

!MetaBenchmark methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:35'!
messageMetaSend
	"self new messageMetaSend" 
	
	|  object meta result |
	object := BenchmarkObject new.
	
	Interpreter onMessageReceivedForObject: object do: [:reifications :reflectogram | 
		reflectogram enable.
		reflectogram returnValue: reflectogram defaultAction value. 
	].
	 
	Interpreter enableMetaBehaviorFor: object.
	
	self assert: (object test = 3).
	
	result := self
		benchmark: [ 1 to: 1000000 do: [ :each | object test ] ]
		times: 10.
	
	Interpreter removeMetaBehaviorOf: object.
	
	^ result! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
experimentalBit
	^ self primitiveGetExperimentalBit ! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
experimentalBit: aBoolean
	self primitiveSetExperimentalBitTo: aBoolean! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/8/2013 15:34'!
identityHash
	"Answer a SmallInteger whose value is related to the receiver's identity.
	This method must not be overridden, except by SmallInteger.
	
	Do not override."

	^ (Interpreter object: self perform: #basicIdentityHash withArguments: {}) bitShift: 18! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:49'!
isExperimentalBitOn
	^ self experimentalBit! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaArgIn: aMessage for: anObject

	^ Interpreter object: anObject metaArgIn: aMessage withReceiver: self! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaReadAt: anIndex

	^ Interpreter object: self metaReadAt: anIndex! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaReceived: aMessage

	^ Interpreter object: self metaReceived: aMessage ! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:39'!
metaRef

	self halt! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaSend: aMessage from: anObject

	^ Interpreter object: anObject metaSend: aMessage withReceiver: self.! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/7/2013 16:07'!
metaWriteAt: anIndex Put: anotherObject

	^ Interpreter object: self metaWriteAt: anIndex Put: anotherObject! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:50'!
primitiveGetExperimentalBit

	<primitive: 150>
	self primitiveFailed! !

!Object methodsFor: '*MetaInterpreter' stamp: 'NickPapoulias 5/5/2013 18:50'!
primitiveSetExperimentalBitTo: aBoolean

	<primitive: 151>
	self primitiveFailed! !

Object subclass: #Reflectogram
	instanceVariableNames: 'trigger returnValue defaultAction'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!Reflectogram class methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:18'!
for: anObject withDefaultAction: aBlock

	^ self basicNew initializeWithTrigger: anObject andAction: aBlock! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:18'!
defaultAction

	^ defaultAction ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
disable 

	Interpreter disableMetaBehaviorFor: trigger ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
enable

	Interpreter enableMetaBehaviorFor: trigger ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:18'!
initializeWithTrigger: anObject andAction: aBlock

	trigger := anObject.
	defaultAction := aBlock.
	! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
isEnabled 

	^ Interpreter isMetaBehaviorEnabledFor: trigger ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
returnValue

	^ returnValue ! !

!Reflectogram methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 6/3/2013 12:19'!
returnValue: aValue

	returnValue := aValue! !

Object subclass: #TestObject
	instanceVariableNames: 'x'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'MetaInterpreter'!

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/11/2013 16:04'!
testInteraction: anObject 

	anObject class.! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/11/2013 15:58'!
testPassingArgument: anObject ! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/9/2013 19:38'!
testSenderHook

	Transcript show: 'this message should be caught'; cr.! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/8/2013 16:01'!
x

	^ x! !

!TestObject methodsFor: 'as yet unclassified' stamp: 'NickPapoulias 5/5/2013 19:02'!
x: anObject 

	x := anObject ! !
Interpreter initialize!
